#include "../StdAfx.h"
#include "../Engine.h"
#include "../Text.h"
#include "../Utils.h"
#include "../Entity.h"
#include "Editor.h"

void Editor::Start() {
}

void Editor::Init() {
    m_level_view.StoreSprite(FT::PlatformTopLeft,        Sprite::GetByName("PlatformTopLeft"));
    m_level_view.StoreSprite(FT::PlatformLeft,           Sprite::GetByName("PlatformLeft"));
    m_level_view.StoreSprite(FT::PlatformMid,            Sprite::GetByName("PlatformMid"));
    m_level_view.StoreSprite(FT::PlatformTop,            Sprite::GetByName("PlatformTop"));
    m_level_view.StoreSprite(FT::PlatformLeftTopRight,   Sprite::GetByName("PlatformLeftTopRight"));
    m_level_view.StoreSprite(FT::PlatformLeftRight,      Sprite::GetByName("PlatformLeftRight"));
    m_level_view.StoreSprite(FT::PlatformTopRight,       Sprite::GetByName("PlatformTopRight"));
    m_level_view.StoreSprite(FT::PlatformRight,          Sprite::GetByName("PlatformRight"));

    m_level_view.StoreSprite(FT::EndOfLevel,             Sprite::GetByName("EndOfLevel"));

    m_level_view.StoreSprite(FT::NcPlatformTopLeft,      Sprite::GetByName("NcPlatformTopLeft"));
    m_level_view.StoreSprite(FT::NcPlatformLeft,         Sprite::GetByName("NcPlatformLeft"));
    m_level_view.StoreSprite(FT::NcPlatformMid,          Sprite::GetByName("NcPlatformMid"));
    m_level_view.StoreSprite(FT::NcPlatformTop,          Sprite::GetByName("NcPlatformTop"));
    m_level_view.StoreSprite(FT::NcPlatformLeftTopRight, Sprite::GetByName("NcPlatformLeftTopRight"));
    m_level_view.StoreSprite(FT::NcPlatformLeftRight,    Sprite::GetByName("NcPlatformLeftRight"));
    m_level_view.StoreSprite(FT::NcPlatformTopRight,     Sprite::GetByName("NcPlatformTopRight"));
    m_level_view.StoreSprite(FT::NcPlatformRight,        Sprite::GetByName("NcPlatformRight"));
}

void Editor::Draw() {
    if (IsInGame()) {
        m_game->Draw();
        return;
    }
    if (IsClearBeforeDraw()) {
        glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);
        glLoadIdentity();
    }

    glPushAttrib(GL_COLOR_BUFFER_BIT);
    {
        glEnable(GL_BLEND);
        glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

        const double tile_width = Engine::Get().GetRenderer()->GetTileWidth();
        const double viewer_x   = -(m_viewer_offset_x * tile_width - 0.45);

        DrawEntitiesPlayerAndLevel(viewer_x);
        DrawBrushAndGui(viewer_x);
    }
    glPopAttrib();

    if (IsSwapAfterDraw()) {
        SDL_GL_SwapBuffers();
    }
}

void Editor::DrawEntitiesPlayerAndLevel(double viewer_x) {
    const double tile_width  = Engine::Get().GetRenderer()->GetTileWidth();
    const double tile_height = Engine::Get().GetRenderer()->GetTileHeight();

    glPushMatrix();
    {
        glTranslated(viewer_x, 0, 0);

        // jednostki i gracz
        std::for_each(m_entities.begin(), m_entities.end(), boost::bind(&Entity::Draw, _1));
        if (m_player_data.x >= 0) {
            const SpritePtr player_sprite = Sprite::GetByName("player_stop");
            const double player_x(m_player_data.x * tile_width), player_y(m_player_data.y * tile_height);
            player_sprite->DrawCurrentFrame(player_x, player_y, tile_width, tile_height);
        } else {
            // nie ma danych gracza, to znaczy, że nie został ustawiony. Dlatego
            // nie rysujemy jego sprite'a. Przy przejściu do podglądu gry zostanie
            // dodany na domyślnej pozycji.
        }
        
        // poziom
        double offset = m_viewer_offset_x;
        m_level_view.SetLevel(m_level, offset);
        m_level_view.Draw(offset);
    }
    glPopMatrix();
}

void Editor::DrawBrushAndGui(double viewer_x) {
    const double tile_width  = Engine::Get().GetRenderer()->GetTileWidth();
    const double tile_height = Engine::Get().GetRenderer()->GetTileHeight();

    glPushAttrib(GL_ENABLE_BIT);
    {
        // wyłącz test głębokości, żeby pędzel oraz GUI były zawsze na wierzchu
        glDisable(GL_DEPTH_TEST);
        BrushPtr default_brush = Brush::New(Sprite::GetByName("PlatformTop"), FT::PlatformTop);
        SetBrush(default_brush);
        if (GetBrush()) {
            glPushMatrix();
            {
                glTranslated(viewer_x, 0, 0);
                Size size(tile_width, tile_height);
                Position position(m_pointer_x * tile_width, m_pointer_y * tile_height);

                Engine::Get().GetRenderer()->DrawQuad(position, position+size, 1,1,1,.4); // podświetlenie
                GetBrush()->GetSprite()->DrawCurrentFrame(position, size);                // obiekt spod pędzla
            }
            glPopMatrix();
        }

        // TODO: tutaj narysować GUI
    }
    glPopAttrib();
}

bool Editor::Update(double dt) {
    if (IsInGame()) {
        m_game->Update(dt);
        if (m_game->IsDone()) {
            SwitchToEditor();
        }
        return !IsDone();
    }

    if (m_keys_down[SDLK_ESCAPE]) {
        SetDone(true);
    } else if (m_keys_down[SDLK_LEFT]) {
        m_viewer_offset_x -= dt * 28.19;
    } else if (m_keys_down[SDLK_RIGHT]) {
        m_viewer_offset_x += dt * 28.24;
    }

    // upewnij się że edytor nie zagląda poza lewą krawędź planszy. Znajdują się tam
    // ujemne wartości na osi odciętych, co może powodować błędy w obliczeniach.
    const double tiles_in_row = 1.0/Engine::Get().GetRenderer()->GetTileWidth();
    m_viewer_offset_x = std::max(m_viewer_offset_x, tiles_in_row/2-1);

    return !IsDone();
}

void Editor::ActionAtCoords(double x, double y) {
    // zawsze stawiaj element planszy
    SetFieldAt(static_cast<size_t>(x), static_cast<size_t>(y), FT::PlatformTop);
}

void Editor::ProcessEvents(const SDL_Event& event) {
    if (IsDone()) {
        return;
    }

    if (event.type == SDL_KEYUP && event.key.keysym.sym==SDLK_0) {
        if (IsInGame()) {
            SwitchToEditor();
        } else {
            LevelPtr level(new Level(m_level, m_entities_to_create, m_player_data));
            level->ShrinkWidth();
            level->SaveEntitiesToFile("data/new.ents");
            level->SaveFieldsToFile("data/new.lvl");
            m_game.reset(new Game(level, PlayerPtr()));
            m_game->Init();
            m_game->Start();
            SwitchToGame();
        }
        return;
    }
    if (IsInGame()) {
        m_game->ProcessEvents(event);
        return;
    }

    if (event.type == SDL_QUIT) {
        SetDone();
    } else if (event.type == SDL_KEYDOWN) {
        m_keys_down[event.key.keysym.sym] = true;
    } else if (event.type == SDL_KEYUP) {
        m_keys_down[event.key.keysym.sym] = false;
    } else if (event.type == SDL_MOUSEMOTION) {
        m_pointer_window_x =       event.motion.x / static_cast<double>(Engine::Get().GetWindow()->GetWidth());
        m_pointer_window_y = 1.0 - event.motion.y / static_cast<double>(Engine::Get().GetWindow()->GetHeight());
        m_pointer_x = MapWindowCoordToWorldX(m_pointer_window_x);
        m_pointer_y = MapWindowCoordToWorldY(m_pointer_window_y);
    } else if (event.type == SDL_MOUSEBUTTONDOWN) {
        m_pointer_window_x =       event.motion.x / static_cast<double>(Engine::Get().GetWindow()->GetWidth());
        m_pointer_window_y = 1.0 - event.motion.y / static_cast<double>(Engine::Get().GetWindow()->GetHeight());
        m_pointer_x = MapWindowCoordToWorldX(m_pointer_window_x);
        m_pointer_y = MapWindowCoordToWorldY(m_pointer_window_y);
        ActionAtCoords(m_pointer_x, m_pointer_y);
    }
}

double Editor::MapWindowCoordToWorldX(double x) const {
    const double tiles_in_row = 1.0/Engine::Get().GetRenderer()->GetTileWidth();
    double k = x*tiles_in_row + m_viewer_offset_x - tiles_in_row/2 + 1;
    return k;
}

double Editor::MapWindowCoordToWorldY(double y) const {
    const double th = Engine::Get().GetRenderer()->GetTileHeight();
    return y/th;
}

void Editor::ClearFieldAt(double x, double y) {
    SetFieldAt(x, y, FT::None);
}

void Editor::SetFieldAt(double x, double y, FT::FieldType ft) {
    m_level->EnsureWidth(static_cast<size_t>(x+1));
    m_level->SetField(static_cast<size_t>(x), static_cast<size_t>(TopDown(y)), ft);
}

FT::FieldType Editor::GetFieldAt(double x, double y) const {
    return m_level->Field(static_cast<size_t>(x), static_cast<size_t>(TopDown(y)));
}
